From 53a87aa7f30bddea27bad4163b1dcf605a1e5307 Mon Sep 17 00:00:00 2001
From: Kevin Wolf <kwolf@redhat.com>
Date: Thu, 13 Nov 2014 12:13:05 +0100
Subject: [PATCH 25/48] Revert "rbd: link and load librbd dynamically"

Message-id: <1415880786-30389-4-git-send-email-kwolf@redhat.com>
Patchwork-id: 62364
O-Subject: [RHEV-7.1 qemu-kvm-rhev PATCH 3/4] Revert "rbd: link and load librbd dynamically"
Bugzilla: 1140744
RH-Acked-by: Paolo Bonzini <pbonzini@redhat.com>
RH-Acked-by: Stefan Hajnoczi <stefanha@redhat.com>
RH-Acked-by: Max Reitz <mreitz@redhat.com>

This reverts commit d46654d71226df0989d21a519bf13ac8680dc06e.

Signed-off-by: Kevin Wolf <kwolf@redhat.com>
Signed-off-by: Miroslav Rezanina <mrezanin@redhat.com>
---
 block/Makefile.objs |   2 +-
 block/rbd.c         | 218 ++++++++++++----------------------------------------
 block/rbd_types.h   |  91 ----------------------
 configure           |  40 +++++++++-
 4 files changed, 90 insertions(+), 261 deletions(-)
 delete mode 100644 block/rbd_types.h

diff --git a/block/Makefile.objs b/block/Makefile.objs
index e078f17..a8258e2 100644
--- a/block/Makefile.objs
+++ b/block/Makefile.objs
@@ -15,7 +15,7 @@ block-obj-y += nbd.o nbd-client.o sheepdog.o
 block-obj-$(CONFIG_LIBISCSI) += iscsi.o
 block-obj-$(CONFIG_LIBNFS) += nfs.o
 block-obj-$(CONFIG_CURL) += curl.o
-block-obj-y += rbd.o
+block-obj-$(CONFIG_RBD) += rbd.o
 block-obj-$(CONFIG_GLUSTERFS) += gluster.o
 block-obj-$(CONFIG_LIBSSH2) += ssh.o
 endif
diff --git a/block/rbd.c b/block/rbd.c
index 8fdd04d..2b797d3 100644
--- a/block/rbd.c
+++ b/block/rbd.c
@@ -11,14 +11,13 @@
  * GNU GPL, version 2 or (at your option) any later version.
  */
 
-#include <gmodule.h>
 #include <inttypes.h>
 
 #include "qemu-common.h"
 #include "qemu/error-report.h"
 #include "block/block_int.h"
 
-#include "rbd_types.h"
+#include <rbd/librbd.h>
 
 /*
  * When specifying the image filename use:
@@ -45,6 +44,13 @@
  * leading "\".
  */
 
+/* rbd_aio_discard added in 0.1.2 */
+#if LIBRBD_VERSION_CODE >= LIBRBD_VERSION(0, 1, 2)
+#define LIBRBD_SUPPORTS_DISCARD
+#else
+#undef LIBRBD_SUPPORTS_DISCARD
+#endif
+
 #define OBJ_MAX_SIZE (1UL << OBJ_DEFAULT_OBJ_ORDER)
 
 #define RBD_MAX_CONF_NAME_SIZE 128
@@ -96,10 +102,6 @@ typedef struct BDRVRBDState {
     char *snap;
 } BDRVRBDState;
 
-static bool librbd_loaded;
-static GModule *librbd_handle;
-
-static int qemu_rbd_load_libs(void);
 static int qemu_rbd_next_tok(char *dst, int dst_len,
                              char *src, char delim,
                              const char *name,
@@ -311,10 +313,6 @@ static int qemu_rbd_create(const char *filename, QemuOpts *opts, Error **errp)
         return -EINVAL;
     }
 
-    if (qemu_rbd_load_libs() < 0) {
-        return -EIO;
-    }
-
     /* Read out options */
     bytes = qemu_opt_get_size_del(opts, BLOCK_OPT_SIZE, 0);
     objsize = qemu_opt_get_size_del(opts, BLOCK_OPT_CLUSTER_SIZE, 0);
@@ -460,10 +458,6 @@ static int qemu_rbd_open(BlockDriverState *bs, QDict *options, int flags,
         goto failed_opts;
     }
 
-    if (qemu_rbd_load_libs() < 0) {
-        return -EIO;
-    }
-
     clientname = qemu_rbd_parse_clientname(conf, clientname_buf);
     r = rados_create(&s->cluster, clientname);
     if (r < 0) {
@@ -592,6 +586,28 @@ static void rbd_finish_aiocb(rbd_completion_t c, RADOSCB *rcb)
     qemu_bh_schedule(acb->bh);
 }
 
+static int rbd_aio_discard_wrapper(rbd_image_t image,
+                                   uint64_t off,
+                                   uint64_t len,
+                                   rbd_completion_t comp)
+{
+#ifdef LIBRBD_SUPPORTS_DISCARD
+    return rbd_aio_discard(image, off, len, comp);
+#else
+    return -ENOTSUP;
+#endif
+}
+
+static int rbd_aio_flush_wrapper(rbd_image_t image,
+                                 rbd_completion_t comp)
+{
+#ifdef LIBRBD_SUPPORTS_AIO_FLUSH
+    return rbd_aio_flush(image, comp);
+#else
+    return -ENOTSUP;
+#endif
+}
+
 static BlockDriverAIOCB *rbd_start_aio(BlockDriverState *bs,
                                        int64_t sector_num,
                                        QEMUIOVector *qiov,
@@ -652,10 +668,10 @@ static BlockDriverAIOCB *rbd_start_aio(BlockDriverState *bs,
         r = rbd_aio_read(s->image, off, size, buf, c);
         break;
     case RBD_AIO_DISCARD:
-        r = rbd_aio_discard(s->image, off, size, c);
+        r = rbd_aio_discard_wrapper(s->image, off, size, c);
         break;
     case RBD_AIO_FLUSH:
-        r = rbd_aio_flush(s->image, c);
+        r = rbd_aio_flush_wrapper(s->image, c);
         break;
     default:
         r = -EINVAL;
@@ -698,6 +714,7 @@ static BlockDriverAIOCB *qemu_rbd_aio_writev(BlockDriverState *bs,
                          RBD_AIO_WRITE);
 }
 
+#ifdef LIBRBD_SUPPORTS_AIO_FLUSH
 static BlockDriverAIOCB *qemu_rbd_aio_flush(BlockDriverState *bs,
                                             BlockDriverCompletionFunc *cb,
                                             void *opaque)
@@ -705,14 +722,19 @@ static BlockDriverAIOCB *qemu_rbd_aio_flush(BlockDriverState *bs,
     return rbd_start_aio(bs, 0, NULL, 0, cb, opaque, RBD_AIO_FLUSH);
 }
 
+#else
+
 static int qemu_rbd_co_flush(BlockDriverState *bs)
 {
+#if LIBRBD_VERSION_CODE >= LIBRBD_VERSION(0, 1, 1)
+    /* rbd_flush added in 0.1.1 */
     BDRVRBDState *s = bs->opaque;
-    if (rbd_flush) {
-        return rbd_flush(s->image);
-    }
+    return rbd_flush(s->image);
+#else
     return 0;
+#endif
 }
+#endif
 
 static int qemu_rbd_getinfo(BlockDriverState *bs, BlockDriverInfo *bdi)
 {
@@ -870,6 +892,7 @@ static int qemu_rbd_snap_list(BlockDriverState *bs,
     return snap_count;
 }
 
+#ifdef LIBRBD_SUPPORTS_DISCARD
 static BlockDriverAIOCB* qemu_rbd_aio_discard(BlockDriverState *bs,
                                               int64_t sector_num,
                                               int nb_sectors,
@@ -879,6 +902,7 @@ static BlockDriverAIOCB* qemu_rbd_aio_discard(BlockDriverState *bs,
     return rbd_start_aio(bs, sector_num, NULL, nb_sectors, cb, opaque,
                          RBD_AIO_DISCARD);
 }
+#endif
 
 static QemuOptsList qemu_rbd_create_opts = {
     .name = "rbd-create-opts",
@@ -914,9 +938,16 @@ static BlockDriver bdrv_rbd = {
 
     .bdrv_aio_readv         = qemu_rbd_aio_readv,
     .bdrv_aio_writev        = qemu_rbd_aio_writev,
+
+#ifdef LIBRBD_SUPPORTS_AIO_FLUSH
     .bdrv_aio_flush         = qemu_rbd_aio_flush,
+#else
     .bdrv_co_flush_to_disk  = qemu_rbd_co_flush,
+#endif
+
+#ifdef LIBRBD_SUPPORTS_DISCARD
     .bdrv_aio_discard       = qemu_rbd_aio_discard,
+#endif
 
     .bdrv_snapshot_create   = qemu_rbd_snap_create,
     .bdrv_snapshot_delete   = qemu_rbd_snap_remove,
@@ -929,153 +960,4 @@ static void bdrv_rbd_init(void)
     bdrv_register(&bdrv_rbd);
 }
 
-typedef struct LibSymbol {
-    const char *name;
-    gpointer *addr;
-} LibSymbol;
-
-static int qemu_rbd_set_functions(GModule *lib, const LibSymbol *funcs)
-{
-    int i = 0;
-    while (funcs[i].name) {
-        const char *name = funcs[i].name;
-        if (!g_module_symbol(lib, name, funcs[i].addr)) {
-            error_report("%s could not be loaded from librbd or librados: %s",
-                         name, g_module_error());
-            return -1;
-        }
-        ++i;
-    }
-    return 0;
-}
-
-/*
- * Set function pointers for basic librados and librbd
- * functions that have always been present in these libraries.
- */
-static int qemu_rbd_set_mandatory_functions(void)
-{
-    LibSymbol symbols[] = {
-        {"rados_create",
-         (gpointer *) &rados_create},
-        {"rados_connect",
-         (gpointer *) &rados_connect},
-        {"rados_shutdown",
-         (gpointer *) &rados_shutdown},
-        {"rados_conf_read_file",
-         (gpointer *) &rados_conf_read_file},
-        {"rados_conf_set",
-         (gpointer *) &rados_conf_set},
-        {"rados_ioctx_create",
-         (gpointer *) &rados_ioctx_create},
-        {"rados_ioctx_destroy",
-         (gpointer *) &rados_ioctx_destroy},
-        {"rbd_create",
-         (gpointer *) &rbd_create},
-        {"rbd_open",
-         (gpointer *) &rbd_open},
-        {"rbd_close",
-         (gpointer *) &rbd_close},
-        {"rbd_resize",
-         (gpointer *) &rbd_resize},
-        {"rbd_stat",
-         (gpointer *) &rbd_stat},
-        {"rbd_snap_list",
-         (gpointer *) &rbd_snap_list},
-        {"rbd_snap_list_end",
-         (gpointer *) &rbd_snap_list_end},
-        {"rbd_snap_create",
-         (gpointer *) &rbd_snap_create},
-        {"rbd_snap_remove",
-         (gpointer *) &rbd_snap_remove},
-        {"rbd_snap_rollback",
-         (gpointer *) &rbd_snap_rollback},
-        {"rbd_aio_write",
-         (gpointer *) &rbd_aio_write},
-        {"rbd_aio_read",
-         (gpointer *) &rbd_aio_read},
-        {"rbd_aio_create_completion",
-         (gpointer *) &rbd_aio_create_completion},
-        {"rbd_aio_get_return_value",
-         (gpointer *) &rbd_aio_get_return_value},
-        {"rbd_aio_release",
-         (gpointer *) &rbd_aio_release},
-        {NULL}
-    };
-
-    if (qemu_rbd_set_functions(librbd_handle, symbols) < 0) {
-        return -1;
-    }
-
-    return 0;
-}
-
-/*
- * Detect whether the installed version of librbd
- * supports newer functionality, and enable or disable
- * it appropriately in bdrv_rbd.
- */
-static void qemu_rbd_set_optional_functions(void)
-{
-    if (g_module_symbol(librbd_handle, "rbd_flush",
-                         (gpointer *) &rbd_flush)) {
-        bdrv_rbd.bdrv_aio_flush = NULL;
-        bdrv_rbd.bdrv_co_flush_to_disk = qemu_rbd_co_flush;
-    } else {
-        rbd_flush = NULL;
-        bdrv_rbd.bdrv_co_flush_to_disk = NULL;
-    }
-
-    if (g_module_symbol(librbd_handle, "rbd_aio_flush",
-                        (gpointer *) &rbd_aio_flush)) {
-        bdrv_rbd.bdrv_co_flush_to_disk = NULL;
-        bdrv_rbd.bdrv_aio_flush = qemu_rbd_aio_flush;
-    } else {
-        rbd_aio_flush = NULL;
-        bdrv_rbd.bdrv_aio_flush = NULL;
-    }
-
-    if (g_module_symbol(librbd_handle, "rbd_aio_discard",
-                        (gpointer *) &rbd_aio_discard)) {
-        bdrv_rbd.bdrv_aio_discard = qemu_rbd_aio_discard;
-    } else {
-        rbd_aio_discard = NULL;
-        bdrv_rbd.bdrv_aio_discard = NULL;
-    }
-}
-
-static int qemu_rbd_load_libs(void)
-{
-    if (librbd_loaded) {
-        return 0;
-    }
-
-    if (!g_module_supported()) {
-        error_report("modules are not supported on this platform: %s",
-                     g_module_error());
-        return -1;
-    }
-
-    librbd_handle = g_module_open("librbd.so.1", 0);
-    if (!librbd_handle) {
-        error_report("error loading librbd: %s", g_module_error());
-        return -1;
-    }
-
-    /*
-     * Due to c++ templates used in librbd/librados and their
-     * dependencies, and linker duplicate trimming rules, closing
-     * librbd would leave it mapped. Make this explicit.
-     */
-    g_module_make_resident(librbd_handle);
-
-    if (qemu_rbd_set_mandatory_functions() < 0) {
-        return -1;
-    }
-    qemu_rbd_set_optional_functions();
-    librbd_loaded = true;
-
-    return 0;
-}
-
 block_init(bdrv_rbd_init);
diff --git a/block/rbd_types.h b/block/rbd_types.h
deleted file mode 100644
index f327cb4..0000000
--- a/block/rbd_types.h
+++ /dev/null
@@ -1,91 +0,0 @@
-/*
- * Types and signatures for librados and librbd
- *
- * Copyright (C) 2013 Inktank Storage Inc.
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, see <http://www.gnu.org/licenses/>.
- */
-
-#ifndef QEMU_BLOCK_RBD_TYPES_H
-#define QEMU_BLOCK_RBD_TYPES_H
-
-/* types from librados used by the rbd block driver */
-
-typedef void *rados_t;
-typedef void *rados_ioctx_t;
-
-static int   (*rados_create)(rados_t *cluster, const char * const id);
-static int   (*rados_connect)(rados_t cluster);
-static void  (*rados_shutdown)(rados_t cluster);
-static int   (*rados_conf_read_file)(rados_t cluster, const char *path);
-static int   (*rados_conf_set)(rados_t cluster, const char *option,
-                               const char *value);
-static int   (*rados_ioctx_create)(rados_t cluster, const char *pool_name,
-                                   rados_ioctx_t *ioctx);
-static void  (*rados_ioctx_destroy)(rados_ioctx_t io);
-
-/* types from librbd used by the rbd block driver*/
-
-typedef void *rbd_image_t;
-typedef void *rbd_completion_t;
-typedef void (*rbd_callback_t)(rbd_completion_t cb, void *arg);
-
-typedef struct {
-    uint64_t id;
-    uint64_t size;
-    const char *name;
-} rbd_snap_info_t;
-
-#define RBD_MAX_IMAGE_NAME_SIZE 96
-#define RBD_MAX_BLOCK_NAME_SIZE 24
-
-typedef struct {
-    uint64_t size;
-    uint64_t obj_size;
-    uint64_t num_objs;
-    int order;
-    char block_name_prefix[RBD_MAX_BLOCK_NAME_SIZE];
-    int64_t parent_pool;
-    char parent_name[RBD_MAX_IMAGE_NAME_SIZE];
-} rbd_image_info_t;
-
-static int      (*rbd_create)(rados_ioctx_t io, const char *name, uint64_t size,
-                              int *order);
-static int      (*rbd_open)(rados_ioctx_t io, const char *name, rbd_image_t *image,
-                            const char *snap_name);
-static int      (*rbd_close)(rbd_image_t image);
-static int      (*rbd_resize)(rbd_image_t image, uint64_t size);
-static int      (*rbd_stat)(rbd_image_t image, rbd_image_info_t *info,
-                            size_t infosize);
-static int      (*rbd_snap_list)(rbd_image_t image, rbd_snap_info_t *snaps,
-                                 int *max_snaps);
-static void     (*rbd_snap_list_end)(rbd_snap_info_t *snaps);
-static int      (*rbd_snap_create)(rbd_image_t image, const char *snapname);
-static int      (*rbd_snap_remove)(rbd_image_t image, const char *snapname);
-static int      (*rbd_snap_rollback)(rbd_image_t image, const char *snapname);
-static int      (*rbd_aio_write)(rbd_image_t image, uint64_t off, size_t len,
-                                 const char *buf, rbd_completion_t c);
-static int      (*rbd_aio_read)(rbd_image_t image, uint64_t off, size_t len,
-                                char *buf, rbd_completion_t c);
-static int      (*rbd_aio_discard)(rbd_image_t image, uint64_t off, uint64_t len,
-                                   rbd_completion_t c);
-static int      (*rbd_aio_create_completion)(void *cb_arg,
-                                             rbd_callback_t complete_cb,
-                                             rbd_completion_t *c);
-static ssize_t  (*rbd_aio_get_return_value)(rbd_completion_t c);
-static void     (*rbd_aio_release)(rbd_completion_t c);
-static int      (*rbd_flush)(rbd_image_t image);
-static int      (*rbd_aio_flush)(rbd_image_t image, rbd_completion_t c);
-
-#endif
diff --git a/configure b/configure
index 3748ccf..d45e1f1 100755
--- a/configure
+++ b/configure
@@ -305,6 +305,7 @@ qom_cast_debug="yes"
 trace_backends="nop"
 trace_file="trace"
 spice=""
+rbd=""
 smartcard_nss=""
 libusb=""
 usb_redir=""
@@ -1031,6 +1032,10 @@ for opt do
   ;;
   --enable-glx) glx="yes"
   ;;
+  --disable-rbd) rbd="no"
+  ;;
+  --enable-rbd) rbd="yes"
+  ;;
   --disable-xfsctl) xfs="no"
   ;;
   --enable-xfsctl) xfs="yes"
@@ -1364,6 +1369,7 @@ Advanced options (experts only):
                            Default:trace-<pid>
   --disable-spice          disable spice
   --enable-spice           enable spice
+  --enable-rbd             enable building the rados block device (rbd)
   --disable-libiscsi       disable iscsi support
   --enable-libiscsi        enable iscsi support
   --disable-libnfs         disable nfs support
@@ -2702,7 +2708,10 @@ if test "$mingw32" = yes; then
 else
     glib_req_ver=2.12
 fi
-glib_modules="gthread-2.0 gmodule-2.0"
+glib_modules=gthread-2.0
+if test "$modules" = yes; then
+    glib_modules="$glib_modules gmodule-2.0"
+fi
 
 for i in $glib_modules; do
     if $pkg_config --atleast-version=$glib_req_ver $i; then
@@ -2843,6 +2852,29 @@ if compile_prog "" "$pthread_lib" ; then
 fi
 
 ##########################################
+# rbd probe
+if test "$rbd" != "no" ; then
+  cat > $TMPC <<EOF
+#include <stdio.h>
+#include <rbd/librbd.h>
+int main(void) {
+    rados_t cluster;
+    rados_create(&cluster, NULL);
+    return 0;
+}
+EOF
+  rbd_libs="-lrbd -lrados"
+  if compile_prog "" "$rbd_libs" ; then
+    rbd=yes
+  else
+    if test "$rbd" = "yes" ; then
+      feature_not_found "rados block device" "Install librbd/ceph devel"
+    fi
+    rbd=no
+  fi
+fi
+
+##########################################
 # libssh2 probe
 min_libssh2_version=1.2.8
 if test "$libssh2" != "no" ; then
@@ -4221,6 +4253,7 @@ echo "spice support     $spice ($spice_protocol_version/$spice_server_version)"
 else
 echo "spice support     $spice"
 fi
+echo "rbd support       $rbd"
 echo "xfsctl support    $xfs"
 echo "nss used          $smartcard_nss"
 echo "libusb            $libusb"
@@ -4615,6 +4648,11 @@ fi
 if test "$qom_cast_debug" = "yes" ; then
   echo "CONFIG_QOM_CAST_DEBUG=y" >> $config_host_mak
 fi
+if test "$rbd" = "yes" ; then
+  echo "CONFIG_RBD=m" >> $config_host_mak
+  echo "RBD_CFLAGS=$rbd_cflags" >> $config_host_mak
+  echo "RBD_LIBS=$rbd_libs" >> $config_host_mak
+fi
 
 echo "CONFIG_COROUTINE_BACKEND=$coroutine" >> $config_host_mak
 if test "$coroutine_pool" = "yes" ; then
-- 
1.8.3.1

